unit CHILDWIN;

interface

uses Windows, Classes, Graphics, Forms, Controls, StdCtrls,
  DXDraws, sysutils, ExtCtrls, Menus, dialogs, ActnList, ImgList, DXClass,
  Grids, ShellAPI, Messages;

const
  TILE_SIZE = 16;
  TILE_SHIFT = 4;
  TILE_COUNT = 190;
  TILE_COLUMNS = 19;
  TILE_ROWS = 10;
  TILE_WIDTH = 304;
  TILE_HEIGHT = 160;
  MAP_X = 1024;
  MAP_Y = 1024;
  DEF_ZOOM = 4;
  //TOOLS
  PEN = 0;         //pen
  ERASE = 1;       //erasor
  SELECT = 2;      //marquee
  DROPPER = 3;     //eye dropper
  STAMPER = 4;     //stamper
  MAGNIFY = 5;     //maginifier
  FILL = 6;        //filler
  FLIPHORI = 7;    //flip horizontal
  FLIPVERT = 8;    //Flip vertical
  LINE = 9;        //Line
  POLYLINE = 10;   //connected line
  RECT = 11;       //rectangle/square
  RECTFILL = 12;   //filled
  CIRCLE = 13;     //ellipse/circle
  CIRCLEFILL = 14; //filled
  AREAFILL = 15;
  
  //tile defs
  //SOLID = [1..161];
  //DOOR = [162..169];
  //FLAG =170;
  //SAFE =171;
  //GOAL =172;
  //FLYOVER = [173..175];
  //FLYUNDER_O = [176..183];
  //FLYUNDER_T = [184..190];

{followed by the bitmapinfoheader}
type

UndoRec = record
  x : array of integer;
  y : array of integer;
  tile : array of byte;
end;

TBitmapInfoHeader = record
  biSize: Longint;   {size of tbitmapinfoheader}
  biWidth: Longint;  {bitmap width}
  biHeight: Longint;  { height of bitmap}
  biPlanes: Word;     {always 1}
  biBitCount: Word;  {number color bits 4 = 16 colors, 8 = 256 pixel is a byte}
  biCompression: Longint; {compression used, 0 }
  biSizeImage: Longint;   {size of the pixel data}
  biXPelsPerMeter: Longint; {not used, 0 }
  biYPelsPerMeter: Longint; {not used, 0 }
  biClrUsed: Longint;       {number of colors used, set to 0 }
  biClrImportant: Longint;  {important colors, set to 0 }
end;

  TMDIChild = class(TDxForm)
    ChildMenu: TMainMenu;
    View1: TMenuItem;
    viewTileWindow: TMenuItem;
    viewSpecialBrushes: TMenuItem;
    N1: TMenuItem;
    viewMapGrid: TMenuItem;
    N2: TMenuItem;
    viewZoomDefault: TMenuItem;
    viewZoomInc: TMenuItem;
    viewZoomDec: TMenuItem;
    Edit: TMenuItem;
    editCut: TMenuItem;
    editCopy: TMenuItem;
    editPaste: TMenuItem;
    N5: TMenuItem;
    editUndo: TMenuItem;
    editRedo: TMenuItem;
    mapAction: TActionList;
    actMapGrid: TAction;
    actZoomIn: TAction;
    actZoomOut: TAction;
    actZoom100: TAction;
    MainToolbar1: TMenuItem;
    StatusBar1: TMenuItem;
    mapVertScroll: TScrollBar;
    pnlBottom: TPanel;
    pnlSpace: TPanel;
    mapHoriScroll: TScrollBar;
    Radar1: TMenuItem;
    Windows: TMenuItem;
    Toolbars: TMenuItem;
    TileInfo1: TMenuItem;
    Tools: TMenuItem;
    Pen1: TMenuItem;
    setErase1: TMenuItem;
    Selection1: TMenuItem;
    setDropper1: TMenuItem;
    Stamp1: TMenuItem;
    Magnifier1: TMenuItem;
    N3: TMenuItem;
    mapDraw: TDXDraw;
    OldTileWindow1: TMenuItem;
    Line1: TMenuItem;
    ConnectedLine1: TMenuItem;
    RectangleSquare1: TMenuItem;
    Clear1: TMenuItem;
    N4: TMenuItem;
    RepLR: TMenuItem;
    RepRL: TMenuItem;
    EraseMap: TMenuItem;
    EllipseCircle1: TMenuItem;
    Fill1: TMenuItem;
    actEditCut: TAction;
    actEditCopy: TAction;
    actEditPaste: TAction;
    actEditRedo: TAction;
    actEditUndo: TAction;
    actEditClear: TAction;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure mapVertScrollChange(Sender: TObject);
    procedure mapHoriScrollChange(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure mapDrawMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure mapDrawMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure placeTile(X, Y : Integer; T : byte);
    procedure placeSpecial(X, Y : integer; S : byte);
    procedure mapDrawMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure mapDrawDblClick(Sender: TObject);
    procedure mapDrawClick(Sender: TObject);
    procedure RepLRClick(Sender: TObject);
    procedure RepRLClick(Sender: TObject);
    procedure EraseMapClick(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure actEditClearExecute(Sender: TObject);
    procedure actEditUndoExecute(Sender: TObject);
    procedure actEditRedoExecute(Sender: TObject);
    procedure actEditCutExecute(Sender: TObject);
  protected
    procedure WMDROPFILES (var Msg: TMessage); message WM_DROPFILES;
  private
    { Private declarations }
    fileName : String;
    //what the tiles are 1..190
    curTileL : byte;
    curTileR : byte;
    usedTile : byte; //the tile that is actually being drawn
    curSpec : byte; //216, 217, 219, 220 never set it to 218
    zoom : byte; //0-5
    Xs, Ys : smallint;//starting point for drawing
    Xe, Ye : smallint;//ending point for drawing
    PolyDraw : boolean; //used in drawing a polyline
    undo : array[0..19] of undorec;
    redo : array[0..19] of undorec;
    float : array of array of byte;
    fX, fY : word;
    hasfloat : boolean;
    procedure resetData();
    procedure DrawLine(const b : boolean);
    procedure DrawRect(const add : boolean);
    procedure DrawCircle(const add : boolean);
    procedure circlePoints(const cx, cy, x, y : integer; const add : boolean);
    procedure fillRight(const x, y : smallint; const old : byte);
    procedure fillDown(const x, y : smallint; const old : byte);
    procedure fillLeft(const x, y : smallint; const old : byte);
    procedure fillUp(const x, y : smallint; const old : byte);
    procedure drawSelect();
    procedure fixsel();
    procedure addUndo(const xpos, ypos : integer; const tileb : byte);
    procedure addRedo(const xpos, ypos : integer; const tileb : byte);
    procedure newUndo();
    procedure newRedo();
    procedure floatsel(Shift: TShiftState);
    procedure defloatsel(Shift: TShiftState);
    procedure setStatus();
    function fillArea(const x, y : smallint; const old : byte) : boolean;
  public
    { Public declarations }
    hasTileset : boolean; //used in saving for whether the map has it's own tileset
    isModified : boolean; //used in closing to determine whether the map has changed
    selDraw : boolean;
    isDrawing : boolean; //whether or not to draw...
    fullMap : array of array of Byte;
    dispX : word; //number of tiles displayed right/left (mapDraw.width/zoom)
    dispY : word; //number of tiles displayed up/down (mapDraw.height/zoom)
    topleftX : word; //top left of map
    topleftY : word;
    tile : TBitmap; //tileset
    mapGrid : boolean; //display a grid over the map
    procedure loadTiles(const name: String);
    procedure LoadMap(const name: String);
    procedure SaveMap(name: String);
    procedure drawMap();
    procedure setLeft(const l : byte);
    procedure setRight(const r : byte);
    procedure setSpecial(const s : byte);
    procedure flipvert();
    procedure fliphori();
    procedure fill();
    procedure rotateCCW();
    procedure rotateCW();
    function zoomIn(): byte;
    function zoomOut() : byte;
    function zoomSet(z : byte; x, y : word) : byte;
    function getFileName() : String;
    function getLeft() : Byte;
    function getRight() : Byte;
  end;

implementation

{$R *.dfm}


uses RADAR,TILE,MAIN, SPECIAL, TILESELECTED;

function TMDIChild.getFileName() : String;
begin
   result := filename;
end;

function tileRect(index: word) : TRect;
begin
//faster method?
case index of
1..19: result:=Bounds((index-1) shl 4, 0, TILE_SIZE,TILE_SIZE);
20..38: result:=Bounds(((index-1)-19) shl 4, 16, TILE_SIZE,TILE_SIZE);
39..57: result:=Bounds(((index-1)-38) shl 4, 32, TILE_SIZE,TILE_SIZE);
58..76: result:=Bounds(((index-1)-57) shl 4, 48, TILE_SIZE,TILE_SIZE);
77..95: result:=Bounds(((index-1)-76) shl 4, 64, TILE_SIZE,TILE_SIZE);
96..114: result:=Bounds(((index-1)-95) shl 4, 80, TILE_SIZE,TILE_SIZE);
115..133: result:=Bounds(((index-1)-114) shl 4, 96, TILE_SIZE,TILE_SIZE);
134..152: result:=Bounds(((index-1)-133) shl 4, 112, TILE_SIZE,TILE_SIZE);
153..171: result:=Bounds(((index-1)-152) shl 4, 128, TILE_SIZE,TILE_SIZE);
172..190: result:=Bounds(((index-1)-171) shl 4, 144, TILE_SIZE,TILE_SIZE);
end;

{
** for use if/when the map format changes
**
**  result := Bounds(TILE_SIZE*((index-1) mod TILE_COLUMNS),
**                     TILE_SIZE*((index-1) div TILE_COLUMNS),
**                     TILE_SIZE, TILE_SIZE)
}
end;

procedure TMDIChild.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if MainForm.MDIChildCount = 1 then
  begin
    TileWindow.Hide;
    RadarBox.Hide;
    TileSel.Hide;
    MainForm.FileSave1.Enabled := False;
    MainForm.FileSaveAs1.Enabled := False;
    MainForm.FileImport1.Enabled := False;
    MainForm.FileClose1.Enabled := False;
  end;
  tile.Free;
  Action := caFree;
end;

procedure TMDIChild.FormCreate(Sender: TObject);
begin
  //makes the array of bytes sized for the map
  DragAcceptFiles(self.Handle, true);
  setlength(Fullmap, MAP_Y, MAP_X);

  //spacing of 4 on each side of the tile box
  //pnlItems.Width := TILE_WIDTH + 8;

  //creating and loading the various bitmaps used
  tile     := TBitmap.Create;

  //at start tiles are default to 1 and 2, and special to 216 ( small asteroid)
  curTileL := 1;
  curTileR := 2;
  curSpec := SpecialBox.spec;

  //loads the default tileset
  loadTiles('');

  //sets up more defaults
  zoom := DEF_ZOOM; //default zoom value
  topLeftX := 0;    //map starts with tile 0,0 at upper left
  topLeftY := 0;
  dispX := 1 + (mapDraw.Width shr zoom);  //how many tiles the map can currently display
  dispY := 1 + (mapDraw.Height shr zoom);

  //map large scrolls are always one screen
  mapVertScroll.LargeChange := dispY-1;
  mapHoriScroll.LargeChange := dispX-1;
  //makes it so that the scrolling fits properly
  mapVertScroll.Max := MAP_Y - (dispY-1);
  mapHoriScroll.Max := MAp_X - (dispX-1);

  fileName := '';
  isModified := False;

  mapGrid := MainForm.ViewMapGrid.Checked;
    case MainForm.tool of
      PEN : mapDraw.Cursor := 16;//can't use 0
      ERASE : mapDraw.Cursor := ERASE;
      DROPPER : mapDraw.Cursor :=  DROPPER;
      STAMPER : mapDraw.Cursor := STAMPER;
      MAGNIFY : mapDraw.Cursor := MAGNIFY;
      SELECT,LINE..CIRCLEFILL : mapDraw.Cursor := SELECT;
      AREAFILL : mapDraw.Cursor := AREAFILL;
    end;
end;

//called whenever the form needs to be repainted
procedure TMDIChild.FormPaint(Sender: TObject);
begin
  drawMap();
end;

{
  used for loading the tileset. and should also be used for
  importing of tilesets. call with '' (empty string) to load the default tiles
}
procedure TMDIChild.loadTiles(const name: String);
var
  bmfFileSize : Longint;
  bmfReserved : Longint;
  bmfBitMapDataOffset : Longint; {from beginning of file}
  BIH : TBitmapInfoHeader;
  f : File;
  idstr:string[2];
  AChar : Char;
  Amt : Integer;
begin
  if not(name = '') then
  begin
    AssignFile(f, name);
    Reset(f, 1);  idstr := '';

    {a bitmap file starts with the id 'BM'}
    BlockRead(f, AChar, 1, amt);
    idstr := Achar;
    BlockRead(f, Achar, 1, amt);
    idstr := idstr + Achar;
    If idstr <> 'BM' then begin
      MessageDlg('The file is not a valid bitmap', mterror, [mbok],0);
      CloseFile(f);
      Exit;
    end;

    {read the file header info}
    BlockRead(f, bmfFileSize, 4, amt);
    BlockRead(f, bmfReserved, 4, amt);
    BlockRead(f, bmfBitMapDataOffset, 4, amt);

    {read the bitmap info header}
    BlockRead(f, BIH.biSize, 4, amt); {size of header itself}
    BlockRead(f, BIH.biWidth, 4, amt);
    BlockRead(f, BIH.biHeight, 4, amt);
    Blockread(f, BIH.biPlanes, 2, amt);
    Blockread(f, BIH.biBitCount, 2, amt);
    BlockRead(f, BIH.biCompression, 4, amt);
    BlockRead(f, BIH.biSizeImage, 4, amt);
    BlockRead(f, BIH.biXPelsPerMeter, 4, amt);
    BlockRead(f, BIH.biYPelsPerMeter, 4, amt);
    BlockRead(f, BIH.biClrUsed, 4, amt);
    BlockRead(f, BIH.biClrImportant, 4, amt);

    {testMemo.Lines.add(idStr + #13#10 +
                       IntToStr(bmfFileSize) + #13#10 +
                       IntToStr(bmfReserved) + #13#10 +
                       IntToStr(bmfBitMapDataOffset));}
    {testMemo.Lines.Add(IntTOStr(BIH.biSize) + #13#10 +
                       IntTOStr(BIH.biWidth) + #13#10 +
                       IntTOStr(BIH.biHeight) + #13#10 +
                       IntTOStr(BIH.biPlanes) + #13#10 +
                       IntTOStr(BIH.biBitCount) + #13#10 +
                       IntTOStr(BIH.biCompression) + #13#10 +
                       IntTOStr(BIH.biSizeImage) + #13#10 +
                       IntTOStr(BIH.biXPelsPerMeter) + #13#10 +
                       IntTOStr(BIH.biYPelsPerMeter) + #13#10 +
                       IntTOStr(BIH.biClrUsed) + #13#10 +
                       IntTOStr(BIH.biClrImportant));}
     CloseFile(f);
     If ((BIH.biWidth = TILE_WIDTH) and (BIH.biHeight = TILE_HEIGHT)) then begin// 304x160
       If BIH.biBitCount <= 8 then //'proper' bmp format
       begin

         tile.LoadFromFile(name);
         hasTileset := true;
         if name <> fileName then
           isModified := true;
         MainForm.FileSave1.Enabled := isModified;

         drawMap();
         TileWindow.Invalidate;
         TileSel.Invalidate;
         exit;
       end
       else//incorrect color depth
         MessageDlg('Incorrect color depth.' + #13#10 + 'Using previous tileset', mtError, [mbOK],0);
     end
     else //incorrect dimensions
       MessageDlg('Incorrect dimensions (' + IntToStr(BIH.biWidth) + ' x ' + IntToStr(BIH.biHeight) + ').' + #13#10 + 'Using previous tileset', mtError, [mbOK],0);
  end
  else
  begin
    tile.LoadFromResourceName(HInstance, 'DEFAULT_TILES_RES');
    drawMap();
    TileWindow.Invalidate;
    TileSel.Invalidate;
  end;

end;

//loading a map weeee fun
procedure TMDIChild.LoadMap(const name: String);
var
InFile: TFileStream;
Dummy: Cardinal;
Word1: Word;
Word2: Word;
Byte1: Byte;
BM : word;
Size: dword;
begin
  filename := name;

  InFile := TFileStream.Create(filename, fmOpenRead or fmShareDenyWrite);

  try
    Infile.readBuffer(BM, 2);  //reads the first two bytes to test for 'BM'
    Infile.readBuffer(Size, 4); //reads in the possible size of a tile bitmap
  finally
    Infile.Free; //file is freed so that the tileset can be loaded
  end;
  if BM = 19778 then LoadTiles(fileName); //if the file starts with a bmp load the tileset


  //actual reading of the file for content is done here
  InFile := TFileStream.Create(fileName, fmOpenReadWrite or fmShareDenyNone);
  if BM = 19778 then
    Infile.Seek(Size,0); //only seek if there was a bmp


  //ADD IN BETTER ERROR CATCHES FOR BAD FILES AND TO HANDLE PURE BMP FILES
  //ADD IN CHECKS TO MAKE SURE X AND Y ARE < 1024 AND TILE [1..190,216..220]


  try
    {
      each tile uses 4 bytes to store data
      all 4 bytes are read into dummy
      then with bitwise operations are divided into their seperate parts
      $FFF =      1111 1111 1111 0000 0000 0000 0000 0000
      $FFF0000 =  0000 0000 0000 1111 1111 1111 0000 0000
      $000000FF = 0000 0000 0000 0000 0000 0000 1111 1111

      so doing a bitwise and on dummy with FFF will return only the first 12 bits

      with FFF0000 it returns the second 12 bits...BUT we then need to shift
      the bits right by 12 so that they make the proper integer

      FF000000 returns only the last 8 bits, and need to be shifted right by 24 bits
    }
    while (InFile.Read(dummy, 4) = 4) do
    begin
      Word1 := Dummy and $FFF;
      Word2 := (Dummy and $FFF000) shr 12;
      Byte1 := (Dummy and $FF000000) shr 24;
      try
        fullMap[word2][word1] := Byte1;
      except
        on eAccessViolation do begin
          if MessageDlg('Error in file continue to try loading?', mtError, [mbYes,mbNo],0) = mrNo then
            exit; end;
      end;
    end;
  finally
    InFile.Free;
  end;
end;

//resizing done for when the form is resized pay attention to order
procedure TMDIChild.FormResize(Sender: TObject);
begin
  resetData();
  Self.Invalidate;
end;

{
  some real fun here drawing the MAP!!!
  it's a simple idea

  we start at the top left tile, and draw as manytiles right as we can
  we draw left to write top to bottom

  the Y value is used on the outer loop because i was told that is more efficient

  when drawing the zoom value is used to determine tile size

}
procedure TMDIChild.drawMap();
var
  curX, curY, startX, startY : integer;
  endX, endY : word;
  curIndex : byte;
begin
  if not mapDraw.CanDraw then exit;

  mapDraw.Surface.Fill(0);
  with mapDraw.Surface.Canvas do
  begin

  //makes sure that everything is drawn without having to do pricey if statements in the loop
  if topLeftX > 4 then startX := -5 else startX := 0;
  if topLeftY > 4 then startY := -5 else startY := 0;
  if (topLeftX + dispX) > 1023 then endX := dispX - 2 else endX := dispX;
  if (topLeftY + dispY) > 1023 then endY := dispY - 2 else endY := dispY;

if mapGrid then
  if zoom > 1 then
  begin
    curindex := 0; //saving some memory
    for curY:= 0 to dispY-1 do
    begin
      if curindex = 5 then
      begin
        Pen.Color:= clsilver;
        curindex := 0;
      end
      else
        Pen.Color:= clGray;
      inc(curindex);           //}
      MoveTo(0, (curY shl zoom) - 1);
      LineTo(mapDraw.Width, (curY shl zoom) - 1);
    end;
    curindex := 0;
    for curX:= 0 to dispX-1 do
    begin
      if curindex = 5 then
      begin
        Pen.Color:= clsilver;
        curindex := 0;
      end
      else
        Pen.Color:= clGray;
      inc(curindex);        //}
      MoveTo((curX shl zoom) - 1, 0);
      LineTo((curX shl zoom) - 1, mapDraw.Height);
    end;
  end;

//seperate routines for drawing the smaller zooms ( tiles look ugly when they get too small )
if zoom > 3 then
begin
  //we begin with -5 so that the stations that are not completely visible still get drawn
  for curY := startY to endY do
    for curX := startX to endX do
      begin
        curIndex := fullMap[curY+topLeftY][curX+TopLeftX]; //gets the currentTile
        case curIndex of
            //NORMAL TILES
          1..190:  CopyRect(bounds(curX shl zoom, curY shl zoom, 1 shl zoom, 1 shl zoom),tile.Canvas,tileRect(curIndex));
            //SMALL ASTEROID
          216,218: StretchDraw(bounds(curX shl zoom, curY shl zoom, 1 shl zoom, 1 shl zoom),SpecialBox.ast_S);
            //LARGE ASTEROID
          217: StretchDraw(bounds(curX shl zoom, curY shl zoom, 2 shl zoom, 2 shl zoom),SpecialBox.ast_L);
            //STATION
          219: StretchDraw(bounds(curX shl zoom, curY shl zoom, 6 shl zoom, 6 shl zoom),SpecialBox.station);
            //WORMHOLE
          220: StretchDraw(bounds(curX shl zoom, curY shl zoom, 5 shl zoom, 5 shl zoom),SpecialBox.wormhole);
        end;
      end;
end
else
begin
  for curY := startY to endY do
    for curX := startX to endX do
      //makes sure the tile is not out of bounds
      begin
        curIndex := fullMap[curY+topLeftY][curX+TopLeftX]; //gets the currentTile
        case curIndex of
        1..161: //solid tiles
          begin
            Brush.Color := clSilver;
            fillRect(bounds(curX shl zoom, curY shl zoom, 1 shl zoom, 1 shl zoom));
          end;
        162..169: //doors
          begin
            Brush.Color := clBlue;
            fillRect(bounds(curX shl zoom, curY shl zoom, 1 shl zoom, 1 shl zoom));
          end;
        170: //flags
          begin
            Brush.Color := clYellow;
            fillRect(bounds(curX shl zoom, curY shl zoom, 1 shl zoom, 1 shl zoom));
          end;
        171: //safe
          begin
            Brush.Color := clLime;
            fillRect(bounds(curX shl zoom, curY shl zoom, 1 shl zoom, 1 shl zoom));
          end;
        172: //goal
          begin
            Brush.Color := clMaroon;
            fillRect(bounds(curX shl zoom, curY shl zoom, 1 shl zoom, 1 shl zoom));
          end;
        173..190: //fly under/over
          begin
            Brush.Color := clAqua;
            fillRect(bounds(curX shl zoom, curY shl zoom, 1 shl zoom, 1 shl zoom));
          end;
        216,218: //SMALL ASTEROID
          begin
            Brush.Color := clGray;
            fillRect(bounds(curX shl zoom, curY shl zoom, 1 shl zoom, 1 shl zoom));
          end;
        217: //LARGE ASTEROID
          begin
            Brush.Color := clGray;
            fillRect(bounds(curX shl zoom, curY shl zoom, 2 shl zoom, 2 shl zoom));
          end;
        219: //STATION
          begin
            Brush.Color := clRed;
            fillRect(bounds(curX shl zoom, curY shl zoom, 6 shl zoom, 6 shl zoom));
          end;
        220: //WORMHOLE
          begin
            Brush.Color := clFuchsia;
            fillRect(bounds(curX shl zoom, curY shl zoom, 5 shl zoom, 5 shl zoom));
          end;
        end;
      end;
end;

if isDrawing then
begin
  case MainForm.Tool of
    LINE : drawLine(false);
    10 : drawLine(false); //POLYLINE = 10 but for some reason is causing errors
    RECT, RECTFILL : drawRect(False);
    CIRCLE, CIRCLEFILL : drawCircle(False);
  end;
end;

if MainForm.Tool = STAMPER then
  if (zoom > 3) then
  begin
    case curSpec of
      216,218:
      begin
        StretchDraw(bounds(Xe, Ye, 1 shl zoom, 1 shl zoom),SpecialBox.ast_S);
        brush.Style := bsDiagCross;
        frameRect(bounds(Xe, Ye, 1 shl zoom, 1 shl zoom));
        Brush.Style := bsSolid;
      end;
      217:
      begin
        StretchDraw(bounds(Xe, Ye, 2 shl zoom, 2 shl zoom),SpecialBox.ast_L);
        brush.Style := bsDiagCross;
        frameRect(bounds(Xe, Ye, 2 shl zoom, 2 shl zoom));
        Brush.Style := bsSolid;
      end;
      219:
      begin
        StretchDraw(bounds(Xe, Ye, 6 shl zoom, 6 shl zoom),SpecialBox.station);
        Brush.Style := bsDiagCross;
        frameRect(bounds(Xe, Ye, 6 shl zoom, 6 shl zoom));
        Brush.Style := bsSolid;
      end;
      220:
      begin
        StretchDraw(bounds(Xe, Ye, 5 shl zoom, 5 shl zoom),SpecialBox.wormhole);
        Brush.Style := bsDiagCross;
        frameRect(bounds(Xe, Ye, 5 shl zoom, 5 shl zoom));
        Brush.Style := bsSolid;
      end;
    end;
  end
  else
    case curSpec of
    216,218: //SMALL ASTEROID
      begin
        Brush.Color := clGray;
        fillRect(bounds(Xe, Ye, 1 shl zoom, 1 shl zoom));
      end;
    217: //LARGE ASTEROID
      begin
        Brush.Color := clGray;
        fillRect(bounds(Xe, Ye, 2 shl zoom, 2 shl zoom));
      end;
    219: //STATION
      begin
        Brush.Color := clRed;
        fillRect(bounds(Xe, Ye, 6 shl zoom, 6 shl zoom));
      end;
    220: //WORMHOLE
      begin
        Brush.Color := clFuchsia;
        fillRect(bounds(Xe, Ye, 5 shl zoom, 5 shl zoom));
      end;
    end;

if ((MainForm.Tool = SELECT) AND selDraw) then
begin
  Pen.Style := psDot;
  drawSelect;
  Pen.Style := psSolid;
end;

Release;
end;
  mapDraw.Flip;
  RadarBox.DrawRadar;
end;

//the scrollbar MAGIC!
procedure TMDIChild.mapVertScrollChange(Sender: TObject);
begin
  topLeftY := mapVertScroll.Position;
  RadarBox.setY(TopLeftY + (dispY shr 1));
  setStatus();
  drawMap();
end;

procedure TMDIChild.mapHoriScrollChange(Sender: TObject);
begin
  topLeftX := mapHoriScroll.Position;
  RadarBox.setX(TopLeftX + (dispX shr 1));
  setStatus();
  drawMap();
end;

procedure TMDIChild.FormActivate(Sender: TObject);
begin
  RadarBox.setChild(self);
  RadarBox.SetX(TopLeftX + (dispX shr 1));
  RadarBox.SetY(TopleftY + (dispY shr 1));
  RadarBox.DrawRadar;

  TileWindow.Invalidate;
  TileSel.Invalidate;

  curSpec := SpecialBox.spec;

  setStatus();
  MainForm.ViewMapGrid.Checked := mapGrid;
  if filename <> '' then
    Mainform.FileSave1.Enabled := isModified
  else
    Mainform.FileSave1.Enabled := false;
end;

//file saving
procedure TMDIChild.saveMap(name: String);
var
  f : File;
  dummy : Cardinal;
  x, y : word;
  written : integer;
begin
  //delete the current file if it exists
  if extractFileExt(name) <> '.lvl' then
    name := name + '.lvl';

  if fileExists(name) then
  begin
    if FileIsReadOnly(name) then
    begin
      MessageDlg('File is read only can not overwrite.', mtError, [mbOK], 0);
      exit;
    end;

    deleteFile(name);
  end;

  if hasTileset then
    begin
    tile.SaveToFile(name);
    assignFile(F, name);
    reset(F,1);
    seek(f,FileSize(F));
    for y := 0 to MAP_Y-1 do
      for x := 0 to MAP_X-1 do
        if fullMap[y][x] <> 0 then
        begin
          Dummy := x or (y shl 12) or ((fullMap[y][x]) shl 24) ; //combines the 3 values into 1 32 bit value
          blockWrite(f, Dummy,4, written);
        end;
    closeFile(F);
    isModified := false;
    MainForm.FileSave1.Enabled := isModified;
  end
  else //did not have a bitmap
  begin
    assignFile(F, name);
    reWrite(F,1);
    for y := 0 to MAP_Y-1 do
      for x := 0 to MAP_X-1 do
        if fullMap[y][x] <> 0 then
        begin
          Dummy := x or (y shl 12) or ((fullMap[y][x]) shl 24) ; //combines the 3 values into 1 32 bit value
          blockWrite(f, Dummy,4, written);
        end;
    closefile(F);
    isModified := false;
    MainForm.FileSave1.Enabled := isModified;
  end;
end;

function TMDIChild.zoomIn() : byte;
var
  x, y : word;
begin
  x := TopLeftX + (dispX shr 1);
  y := TopLeftY + (dispY shr 1);
  result := zoomSet(zoom + 1, x, y);
end;

function TMDIChild.zoomOut() : byte;
var
  x, y : word;
begin
  x := TopLeftX + (dispX shr 1);
  y := TopLeftY + (dispY shr 1);
  result := zoomSet(zoom - 1, x, y);
end;

function TMDIChild.zoomSet(z : byte; x, y : word) : byte;
begin
  if x = 0 then x := TopLeftX + (dispX shr 1);
  if y = 0 then y := TopLeftY + (dispY shr 1);
  if (z <= 5) then
  begin
    zoom := z;
    resetData();

    if x < (dispX shr 1) then
      TopLeftX := TopLeftX
    else if x > (1023 - (dispX shr 1)) then
      topleftX := 1023 - dispX
    else
      topLeftX := x - (dispX shr 1);

    if y < (dispY shr 1) then
      TopLeftY := TopLeftY
    else if y > (1023 - (dispY shr 1)) then
      topleftY := 1023 - dispY
    else
      topLeftY := y - (dispY shr 1);

    MapVertScroll.Position := TopLeftY;
    MapHoriScroll.Position := TopLeftX;
    if RadarBox.Visible then
    begin
      RadarBox.setX(TopLeftX + (DispX shr 1));
      RadarBox.setY(TopLeftY + (DispY shr 1));
      RadarBox.DrawRadar;
    end;

    setStatus();
    drawMap();
  end;
  result := zoom;
end;

{
  resizes items properly and gets new data whenever things are changed
  should be for when:
  things resize
  zooming
  scrolls
}
procedure TMDIChild.resetData();
begin
  dispX := 1 + (mapDraw.Width shr zoom);
  dispY := 1 + (mapDraw.Height shr zoom);
  if dispX > 1023 then dispX := 1023;
  if dispY > 1023 then dispY := 1023;
  mapVertScroll.LargeChange := dispY-1;
  mapHoriScroll.LargeChange := dispX-1;
  mapVertScroll.Max         := MAP_Y - (dispY-1);
  mapHoriScroll.Max         := MAp_X - (dispX-1);
  setStatus();
end;

function TMDIChild.getLeft() : byte;
begin
  result := CurTileL;
end;

function TMDIChild.getRight() : Byte;
begin
  result := CurTileR;
end;

procedure TMDIChild.setLeft(const l : byte);
begin
  curTileL := l;
end;

procedure TMDIChild.setRight(const r : byte);
begin
  curTileR := r;
end;

procedure TMDIChild.setSpecial(const s: Byte);
begin
  curSpec := s;
end;

procedure TMDIChild.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  if isModified then
    case MessageDlg('Changes have been made would you like to save the file ' + fileName, mtWarning, [mbYes,mbNo,mbCancel], 0) of
      mrCancel : canClose := false;
      mrNo : canClose := true;
      mrYes :
        begin
          if fileName = '' then
          begin
            if isModified then
              if MainForm.SaveDialog.Execute then
              begin
                SaveMap(MainForm.saveDialog.FileName);
                //MainForm.MRU.AddHistoryFile(MainForm.saveDialog.FileName); //removed until reinstalled
              end
              else
                canClose := false;
          end
          else
          begin
            SaveMap(FileName);
            //MainForm.MRU.AddHistoryFile(fileName); //removed until reinstalled
          end;
        end;
    end;
end;

procedure TMDIChild.placeTile(X, Y : integer; T : byte);
var
 xtest, ytest: word;
begin

if ((X < 1024) and (Y < 1024)) then
  if ((X >= 0) and (y >= 0)) then
  begin
    //testing for the station
    if ((x < 5) and (y < 5)) then
    begin
      for ytest := 0 to y do
        for xtest := 0 to x do
          if fullmap[ytest][xtest] = 219 then
          begin
            addUndo(xtest,ytest,fullMap[ytest][xtest]);
            fullmap[ytest][xtest] := 0;
          end;
    end
    else if ((x < 5) and (y >= 5)) then
    begin
      for ytest := y-5 to y do
        for xtest := 0 to x do
          if fullmap[ytest][xtest] = 219 then
          begin
            addUndo(xtest,ytest,fullMap[ytest][xtest]);
            fullmap[ytest][xtest] := 0;
          end;
    end
    else if ((y < 5) and (x >= 5)) then
    begin
      for ytest := 0 to y do
        for xtest := x-5 to x do
          if fullmap[ytest][xtest] = 219 then
          begin
            addUndo(xtest,ytest,fullMap[ytest][xtest]);
            fullmap[ytest][xtest] := 0;
          end;
    end
    else
      for ytest := y-5 to y do
        for xtest := x-5 to x do
          if fullmap[ytest][xtest] = 219 then
          begin
            addUndo(xtest,ytest,fullMap[ytest][xtest]);
            fullmap[ytest][xtest] := 0;
          end;

    //testing for the wormhole
    if ((x < 4) and (y < 4)) then
    begin
      for ytest := 0 to y do
        for xtest := 0 to x do
          if fullmap[ytest][xtest] = 220 then
          begin
            addUndo(xtest,ytest,fullMap[ytest][xtest]);
            fullmap[ytest][xtest] := 0;
          end;
    end
    else if ((x < 4) and (y >= 4)) then
    begin
      for ytest := y-4 to y do
        for xtest := 0 to x do
          if fullmap[ytest][xtest] = 220 then
          begin
            addUndo(xtest,ytest,fullMap[ytest][xtest]);
            fullmap[ytest][xtest] := 0;
          end;
    end
    else if ((y < 4) and (x >= 4)) then
    begin
      for ytest := 0 to y do
        for xtest := x-4 to x do
          if fullmap[ytest][xtest] = 220 then
          begin
            addUndo(xtest,ytest,fullMap[ytest][xtest]);
            fullmap[ytest][xtest] := 0;
          end;
    end
    else
      for ytest := y-4 to y do
        for xtest := x-4 to x do
          if fullmap[ytest][xtest] = 220 then
          begin
            addUndo(xtest,ytest,fullMap[ytest][xtest]);
            fullmap[ytest][xtest] := 0;
          end;

    //woohoo screwy logic :) testing for asteroid
    if x > 0 then
    begin
      if fullMap[y][x-1] = 217 then
      begin
            addUndo(x-1,y,fullMap[y][x-1]);
            fullMap[y][x-1] := 0;
      end; 
      if y > 0 then
      begin
        if fullMap[y-1][x] = 217 then
        begin
            addUndo(x,y-1,fullMap[y-1][x]);
            fullMap[y-1][x] := 0;
        end;
        if fullMap[y-1][x-1] = 217 then
        begin
            addUndo(x-1,y-1,fullMap[y-1][x-1]);
            fullMap[y-1][x-1] := 0;
        end;
      end;
    end
    else if y > 0 then
    begin
      if fullMap[y-1][x] = 217 then
      begin
            addUndo(x,y-1,fullMap[y-1][x]);
            fullMap[y-1][x] := 0;
      end;
      if x > 0 then
      begin
        if fullMap[y][x-1] = 217 then
        begin
            addUndo(x-1,y,fullMap[y][x-1]);
            fullMap[y][x-1] := 0;
        end;
        if fullMap[y-1][x-1] = 217 then
        begin
            addUndo(x-1,y-1,fullMap[y-1][x-1]);
            fullMap[y-1][x-1] := 0;
        end;
      end;
    end;

    if T <> fullMap[y][x] then
      addUndo(x,y,fullMap[y][x]);

    fullMap[y][x] := T;
  end;
end;

procedure TMDIChild.placeSpecial(X, Y : integer; S : byte);
var
a,b : word;
begin
  case S of
  216,218: placeTile(x, y, s);
  217:
    begin
    if x > 1022 then x := 1022;
    if y > 1022 then y := 1022;
      for a := y to y+1 do
        for b := x to x+1 do
          placeTile(b,a,0);
      placeTile(x,y,s);
    end;
  219:
    begin
    if x > 1018 then x := 1018;
    if y > 1018 then y := 1018;
      for a := y to y+5 do
        for b := x to x+5 do
          placeTile(b,a,0);
      placeTile(x,y,s);
    end;
  220:
    begin
    if x > 1019 then x := 1019;
    if y > 1019 then y := 1019;
      for a := y to y+4 do
        for b := x to x+4 do
          placeTile(b,a,0);
      placeTile(x,y,s);
    end;
  end;
end;
{
-------------------------------------------------------------------------------|
                                                                               |
                                                                               |
                                                                               |
                                                            easier to notice   |
                                                            when scrolling     |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
                                                                               |
-------------------------------------------------------------------------------|
}
procedure TMDIChild.mapDrawMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  xcoord, ycoord : integer;
begin
  if MainForm.Tool = MAGNIFY then //gonna try and make all these ifs easier to follow there has to be a way to eliminate thse
  begin
    if  button = mbLeft then
      zoomSet(zoom+1, (x shr zoom)+topLeftX,(y shr zoom) + topLeftY)
    else
      zoomSet(zoom-1, (x shr zoom)+topLeftX,(y shr zoom) + topLeftY);
  end
  else if MainForm.Tool = DROPPER then
  begin
    xcoord := (X shr zoom) + topLeftX;
    ycoord := (Y shr zoom) + topLeftY;
    if ((xcoord < 1024) and (ycoord < 1024)) then
      if fullmap[ycoord][xcoord] in [1..190] then
        case Button of
          mbLeft : curTileL := fullmap[ycoord][xcoord];
          mbRight : curTileR := fullmap[ycoord][xcoord];
        end;
    TileSel.Invalidate;
    TileWindow.Invalidate;
  end else begin  //if it is not the magnifier or the eye dropper
    case Button of //figure out what button was pushed used for all other tools
      mbLeft: usedTile := curTileL;
      mbMiddle : usedTile := 0;
      mbRight: usedTile := curTileR;
    end;

    if hasfloat = true then //if there is a floating selection
    begin
      xcoord := (X shr zoom)+topLeftX;
      ycoord := (Y shr zoom)+topLeftY;
      //if the mouse press is outside the current selection defloat it.
      if not (ycoord in [ys..ye]) then
        if not (xcoord in [xs..xe]) then
          defloatsel(Shift);
    end
    else if selDraw = true then //if there is a selection but it's not floating
    begin
      xcoord := (X shr zoom)+topLeftX;
      ycoord := (Y shr zoom)+topLeftY;

      if ((ycoord in [ys..ye]) and (xcoord in [xs..xe])) then //if clicking inside the selection
      begin
        floatsel(Shift);
      end else
      begin  //clicking outside the selection i'm confused about this...
        Xe := (X shr zoom)+topLeftX;
        Ye := (Y shr zoom)+topLeftY;

        if isDrawing then
          if MainForm.Tool = POLYLINE then
            drawLine(true);

        Xs := Xe;
        Ys := Ye;
      end;//
    end // end seldraw
    else
    begin //no float and no selection used for polyine drawing.
      Xe := (X shr zoom)+topLeftX;
      Ye := (Y shr zoom)+topLeftY;

      if isDrawing then
        if MainForm.Tool = POLYLINE then
          drawLine(true);

      Xs := Xe;
      Ys := Ye;
    end;

  if MainForm.Tool <> POLYLINE then //resets variables, not used with polyline
  begin
    isDrawing := true; //currently drawing
    isModified := true; //the file has been modified
    if filename <> '' then //if it's not a new file enable saving
      Mainform.FileSave1.Enabled := isModified
    else
      Mainform.FileSave1.Enabled := false;
    if MainForm.Tool <> SELECT then //start recording a new undo
      newundo();
  end;

  case MainForm.Tool of
    PEN :
      if ssCtrl in Shift then //if ctrl is being held act as an erasor
        placeTile(Xs, Ys, 0)
      else
        //fillArea((X shr Zoom) + TopLeftX, (Y shr Zoom) + topLeftY, usedTile, fullmap[(Y shr Zoom) + topLeftY][(X shr Zoom) + TopLeftX]);
        placeTile(Xs,Ys, usedTile);
    ERASE : placeTile(Xs,Ys, 0);
    STAMPER : placeSpecial(Xs, Ys, curSpec);
    SELECT:
      begin
        if hasfloat = true then //if there is a floating selection
        begin

        end
        else if selDraw = true then //if there is already a selection present
        begin
          //if
        end
        else begin //making a new selection
          selDraw := true; //yes there is a selection to draw
          MainForm.selFill.Enabled := true; //enable selection fill
          MainForm.selFlipHori.Enabled := true; //enable flip horizontal
          Mainform.selFlipVert.Enabled := true; //enable flip vertical
          actEditClear.Enabled := true;
          actEditCut.Enabled := true;
        end;
      end;
    AREAFILL: //the area fill has problems if the size is too large.
      begin
        saveMap('~CLEbackup.lvl'); //a backup is created incase the stack overflows
        if fillArea(Xs,Ys,fullmap[Ys][Xs]) then
          if FileExists('~CLEbackup.lvl') then //if the program didn't crash then the backup is removed
             DeleteFile('~CLEbackup.lvl');
      end;
  end;

  end;
  DrawMap();
end;

procedure TMDIChild.mapDrawMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if MainForm.Tool <> POLYLINE then
    isDrawing := false;

  case MainForm.Tool of //actually add the shape to the map on release
    LINE : drawline(true);
    RECT, RECTFILL : drawRect(true);
    CIRCLE, CIRCLEFILL : drawCircle(true);
    SELECT : fixSel();
  end;
  drawMap();
end;

procedure TMDIChild.mapDrawMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var
  Dx, Dy : integer;
  Xss, Yss : word;
begin
  //updates for the status bar
  Xss := (X shr zoom) + topLeftX;
  Yss := (Y shr zoom) + topLeftY;

  Mainform.StatusBar.Panels.Items[1].Text := InttoStr(Xss) + ', ' + InttoStr(Yss) + ' :: ';
  Mainform.StatusBar.Panels.Items[1].Text := Mainform.StatusBar.Panels.Items[1].Text + Char(byte('A') + (Xss *20 div 1024));
  Mainform.StatusBar.Panels.Items[1].Text := Mainform.StatusBar.Panels.Items[1].Text + InttoStr(Yss *20 div 1024 + 1);

  //if something is being drawn
  if isDrawing then
  begin
    case MainForm.Tool of
      PEN :
        begin  //pen and erasor are pretty simple to understand
          if ssCtrl in shift then
            placeTile((X shr zoom) + topLeftX, (Y shr zoom) + topLeftY, 0)
          else
            placeTile((X shr zoom) + topLeftX, (Y shr zoom) + topLeftY, usedTile);
          DrawMap();
        end;
      ERASE :
        begin
          placeTile((X shr zoom) + topLeftX, (Y shr zoom) + topLeftY, 0);
          DrawMap();
        end;
      LINE,POLYLINE:
        begin  //makes sure that the end has the proper sign when the mouse goes off the canvas
          if X >= 0 then Xe := (X shr zoom) + TopLeftX
          else Xe := -(abs(X) shr zoom) + TopLeftX;
          if Y >= 0 then Ye := (Y shr zoom) + TopLeftY
          else Ye := -(abs(Y) shr zoom) + TopLeftY;
          mainForm.StatusBar.Panels.Items[0].Text :=    '(' + inttostr(Xs) + ',' + inttoStr(Ys) + ')' + ' - ' +
                                                        '(' + inttostr(Xe) + ',' + inttoStr(Ye) + ')'
                                                        + ' :: ' + inttostr(abs(Xs-Xe)+1) + ',' + inttostr(abs(Ys-Ye)+1);

          if ssShift in shift then //snap if shift is held...the snapping code is kinda wierd
          begin
            dY := abs(Ys - Ye);
            dX := abs(Xs - Xe);

            if Xe <= Xs then  //left
            begin
              if dX > dY then
              begin
                if Ye <= Ys then //left up
                begin
                  if Ye > (Ys -(dx shr 1)) then Ye := Ys
                  else Ye := Ys - Dx;
                end else begin //left down
                  if Ye < (Ys +(dx shr 1)) then Ye := Ys
                  else Ye := Ys + Dx;
                end;
              end else begin
                if Xe > (Xs - (Dy shr 1)) then
                  Xe := Xs
                else
                  Xe := Xs - Dy;
              end;
            end
            else
            begin
              if dX > dY then
              begin
                if Ye <= Ys then //right up
                begin
                  if Ye > (Ys -(dx shr 1)) then Ye := Ys
                  else Ye := Ys - Dx;
                end else begin
                  if Ye < (Ys +(dx shr 1)) then Ye := Ys
                  else Ye := Ys + Dx;
                end;
              end else begin
                if Xe < (Xs + (Dy shr 1)) then
                  Xe := Xs
                else
                  Xe := Xs + Dy;
              end;
            end;
          end;// if ssShift in shift then

          drawMap;
        end;
      RECT, RECTFILL, CIRCLE, CIRCLEFILL:
        begin //same as with the line
          if X >= 0 then Xe := (X shr zoom) + TopLeftX
          else Xe := -(abs(X) shr zoom) + TopLeftX;
          if Y >= 0 then Ye := (Y shr zoom) + TopLeftY
          else Ye := -(abs(Y) shr zoom) + TopLeftY;
          mainForm.StatusBar.Panels.Items[0].Text :=    '(' + inttostr(Xs) + ',' + inttoStr(Ys) + ')' + ' - ' +
                                                        '(' + inttostr(Xe) + ',' + inttoStr(Ye) + ')'
                                                        + ' :: ' + inttostr(abs(Xs-Xe)+1) + ',' + inttostr(abs(Ys-Ye)+1);

          if ssShift in shift then //snapping code again
          begin
            //snap to a square
            dY := abs(Ys - Ye);
            dX := abs(Xs - Xe);
                        if ((Xs - Xe) > 0) then
            begin
              if ((Ys - Ye) > 0) then
              begin
                if dX > dY then Ye := Ys + (Xe - Xs)
                else            Xe := Xs + (Ye - Ys);
              end else begin
                if dX > dY then Ye := Ys + (Xs - Xe)
                else            Xe := Xs + (Ys - Ye);
              end;
            end else
              if ((Ys - Ye) > 0) then
              begin
                if dX > dY then Ye := Ys + (Xs - Xe)
                else            Xe := Xs + (Ys - Ye);
              end else begin
                if dX > dY then Ye := Ys + (Xe - Xs)
                else            Xe := Xs + (Ye - Ys);
              end;
          end;

          drawMap;
        end;
      SELECT: //might need to be cleaned up
        begin
          //coordinates are based on map
          if (X < 0) then
            Xe := 0
          else if (X > 1023) then
            Xe := 1023
          else
            Xe := (X shr zoom) + TopLeftX;

          if (Y < 0 ) then
            Ye :=0
          else if (Y > 1023) then
            Ye := 1023
          else
            Ye := (Y shr zoom) + TopLeftY;
          mainForm.StatusBar.Panels.Items[0].Text :=    '(' + inttostr(Xs) + ',' + inttoStr(Ys) + ')' + ' - ' +
                                                        '(' + inttostr(Xe) + ',' + inttoStr(Ye) + ')'
                                                        + ' :: ' + inttostr(abs(Xs-Xe)+1) + ',' + inttostr(abs(Ys-Ye)+1);

          if ssShift in shift then
          begin
            //snap to a square
            dY := abs(Ys - Ye);
            dX := abs(Xs - Xe);

            if ((Xs - Xe) > 0) then
            begin
              if ((Ys - Ye) > 0) then
              begin
                if dX > dY then Ye := Ys + (Xe - Xs)
                else            Xe := Xs + (Ye - Ys);
              end else begin
                if dX > dY then Ye := Ys + (Xs - Xe)
                else            Xe := Xs + (Ys - Ye);
              end;
            end else
              if ((Ys - Ye) > 0) then
              begin
                if dX > dY then Ye := Ys + (Xs - Xe)
                else            Xe := Xs + (Ys - Ye);
              end else begin
                if dX > dY then Ye := Ys + (Xe - Xs)
                else            Xe := Xs + (Ye - Ys);
              end;
          if (Xe < 0) then
            Xe := 0
          else if (Xe > 1023) then
            Xe := 1023;

          if (Ye < 0 ) then
            Ye :=0
          else if (Ye > 1023) then
            Ye := 1023;
          end;

          drawMap;
        end;
    end;
  end;

  if not isDrawing then
  if MainForm.Tool = POLYLINE then// the poly line is special drawing it doesn't start until the mouse is released the first time
    if PolyDraw then
    begin
      Xe := (X shr zoom) + TopleftX;
      Ye := (Y shr zoom) + TopLeftY;
      Xs := Xe;
      Ys := Ye;
      isDrawing := true;
      isModified := true;
    end;

  if MainForm.Tool = STAMPER then //stamps are placed at mouse release
  begin
    Xe := (X shr zoom) shl zoom;//snaps it to grid
    Ye := (Y shr zoom) shl zoom;
    drawmap;
  end;

end;

procedure TMDIChild.DrawLine(const b : boolean);
var
  den, num, numadd, numpixels,
  xinc1, xinc2, yinc1, yinc2,
  x, y, deltax, deltay : integer;
  curpixel : integer;
begin
  deltax := abs(Xe - Xs);        // The difference between the x's
  deltay := abs(Ye - Ys);        // The difference between the y's
  if XS > -1 then x := Xs-topleftX
  else x := Xs + topLeftX;                       // Start x off at the first pixel
  if Ys > -1 then y := Ys-topleftY             // Start y off at the first pixel
  else y := Ys + topLeftY;

  if (Xe >= Xs) then                 // The x-values are increasing
  begin
    xinc1 := 1;
    xinc2 := 1;
  end
  else                          // The x-values are decreasing
  begin
    xinc1 := -1;
    xinc2 := -1
  end;

  if (Ye >= Ys) then                 // The y-values are increasing
  begin
    yinc1 := 1;
    yinc2 := 1;
  end
  else                          // The y-values are decreasing
  begin
    yinc1 := -1;
    yinc2 := -1;
  end;

  if (deltax >= deltay) then         // There is at least one x-value for every y-value
  begin
    xinc1 := 0;                  // Don't change the x when numerator >= denominator
    yinc2 := 0;                  // Don't change the y for every iteration
    den := deltax;
    num := deltax shr 1;
    numadd := deltay;
    numpixels := deltax;         // There are more x-values than y-values
  end
  else                          // There is at least one y-value for every x-value
  begin
    xinc2 := 0;                  // Don't change the x for every iteration
    yinc1 := 0;                  // Don't change the y when numerator >= denominator
    den := deltay;
    num := deltay shr 1;
    numadd := deltax;
    numpixels := deltay;         // There are more y-values than x-values
  end;

  for curpixel := 0 to numpixels do
  begin
    if (zoom > 3) then
      if usedTile <> 0 then
        mapDraw.Surface.Canvas.CopyRect(bounds(x shl zoom, y shl zoom, 1 shl zoom, 1 shl zoom),
                 tile.Canvas,tileRect(usedTile))
      else
      begin
        mapDraw.Surface.Canvas.Brush.Color := clBlack;
        mapDraw.Surface.Canvas.fillRect(bounds(X shl zoom, Y shl zoom, 1 shl zoom, 1 shl zoom));
      end
    else
    begin
      case usedTile of
        0: mapDraw.Surface.Canvas.Brush.Color := clBlack;
        1..161: mapDraw.Surface.Canvas.Brush.Color := clSilver;
        162..169: mapDraw.Surface.Canvas.Brush.Color := clBlue;
        170: mapDraw.Surface.Canvas.Brush.Color := clYellow;
        171: mapDraw.Surface.Canvas.Brush.Color := clLime;
        172: mapDraw.Surface.Canvas.Brush.Color := clMaroon;
        173..190: mapDraw.Surface.Canvas.Brush.Color := clAqua;
      end;
      mapDraw.Surface.Canvas.fillRect(bounds(X shl zoom, Y shl zoom, 1 shl zoom, 1 shl zoom));
    end;

    if b then //only add the tile to the map when mouse is released
      placeTile(x+topLeftX,y+topleftY, usedTile);

    inc(num,numadd);              // Increase the numerator by the top of the fraction
    if (num >= den) then          // Check if numerator >= denominator
    begin
      dec(num,den);               // Calculate the new numerator value
      inc(x,xinc1);               // Change the x as appropriate
      inc(y,yinc1);               // Change the y as appropriate
    end;
    inc(x,xinc2);                 // Change the x as appropriate
    inc(y,yinc2);                 // Change the y as appropriate
  end;
end;

procedure TMDIChild.DrawRect(const add : boolean);
var
  x1, x2, y1, y2 : integer;
begin
  x1 := Xs; //backup storage for the start and stop coords
  x2 := Xe;
  y1 := Ys;
  y2 := Ye;

    //to draw a rectangle you draw 4 lines weeeeeee
    //drawline draws from (Xs, Ys) to (Xe, Ye)
    //makes sure Xe and Xs are the same ie a vertical line
    Xe := x1;
    drawline(add);//left side
    Xe := x2;
    Ye := y1;
    drawLine(add);//top
    Xs := x2;
    Ys := y2;
    Xe := x1;
    Ye := y2;
    drawLine(add);//bottom
    Xe := x2;
    Ye := y1;
    drawLine(add);//right i think i may have these wrong but they work so shouldnt be changed

  //reload the start and stop coords
  Xs := x1;
  Ys := y1;
  Ye := Y2;
  Xe := x2;
end;

procedure TMDIChild.drawSelect();
var
  x, y, x2, y2 : word;
  selarea : array[0..4] of TPoint;
begin
   x := xs;
   x2 := xe;
   y := ys;
   y2 := ye;


  if Xe < Xs then
  begin
    x2 := Xs;
    x := Xe;
  end;
  if Ye < Ys then
  begin
    y := Ye;
    y2 := Ys;
  end;


  //draws a selection rectangle
  //creates an array of 4 points then draws a line between those points

  selarea[0].X := (X - TopLeftX) shl zoom;
  selarea[0].Y := (Y - TopLeftY) shl zoom;

  selarea[1].X := (X2 - TopLeftX) shl zoom +15;//adds 15 to be on the outside edge of the tiles
  selarea[1].Y := (Y - TopLeftY) shl zoom;

  selarea[2].X := (X2 - TopLeftX) shl zoom +15;
  selarea[2].Y := (Y2 - TopLeftY) shl zoom +15;

  selarea[3].X := (X - TopLeftX) shl zoom;
  selarea[3].Y := (Y2 - TopLeftY) shl zoom +15;

  selarea[4].X := (X - TopLeftX) shl zoom;
  selarea[4].Y := (Y - TopLeftY) shl zoom;

  MapDraw.Surface.Canvas.Polyline(selarea);

end;

procedure TMDIChild.mapDrawDblClick(Sender: TObject);
begin
  //only used with the polyline tool
  if (isDrawing and (MainForm.Tool = POLYLINE)) then
  begin
    isDrawing := false;
    drawLine(true);
    polyDraw := false;
    drawMap;
  end;

end;

procedure TMDIChild.mapDrawClick(Sender: TObject);
begin
  //only used with polyline
  if ((not isDrawing) and (MainForm.Tool = POLYLINE)) then
  begin
    polyDraw := true;
    newUndo();
  end;
end;

procedure TMDIChild.fill();//fill selection simple overwrites anything in the selection
var
  x, y : integer;
begin

  if selDraw then
  begin
    for y := Ys to Ye do
      for x := Xs to Xe do
        placetile(x, y, usedTile);

   drawMap;
  end;

end;

procedure TMDIChild.flipvert();
var
  xStart, yStart : word;
  temp : byte;
begin
  if selDraw then
  begin
   for xstart := xs to xe do
     for ystart := yS to (ys + ((ye - ys) shr 1)) do
       begin
         temp := fullmap[ystart][xstart];
         fullmap[ystart][xstart] := fullmap[ye-(ystart-ys)][xstart];
         fullmap[ye-(ystart-ys)][xstart] := temp;
       end;
   for ystart := ys to ye do
     for xstart := xS to xe do
       case fullmap[ystart][xstart] of
         217: begin
           fullmap[ystart-1][xstart] := fullmap[ystart][xstart];
           fullmap[ystart][xstart] := 0; end;
         219: begin
           fullmap[ystart-5][xstart] := fullmap[ystart][xstart];
           fullmap[ystart][xstart] := 0; end;
         220: begin
           fullmap[ystart-4][xstart] := fullmap[ystart][xstart];
           fullmap[ystart][xstart] := 0; end;
       end;

    drawMap;
  end;
end;

procedure TMDIChild.fliphori();
var
  xStart, yStart : word;
  temp : byte;
begin
  if selDraw then
  begin
   for ystart := ys to ye do
     for xstart := xS to (xs + ((xe - xs) shr 1)) do
       begin
         temp := fullmap[ystart][xstart];
         fullmap[ystart][xstart] := fullmap[ystart][xe-(xstart-xs)];
         fullmap[ystart][xe-(xstart-xs)] := temp;
       end;
   for ystart := ys to ye do
     for xstart := xS to xe do
       case fullmap[ystart][xstart] of
         217: begin
           fullmap[ystart][xstart-1] := fullmap[ystart][xstart];
           fullmap[ystart][xstart] := 0; end;
         219: begin
           fullmap[ystart][xstart-5] := fullmap[ystart][xstart];
           fullmap[ystart][xstart] := 0; end;
         220: begin
           fullmap[ystart][xstart-4] := fullmap[ystart][xstart];
           fullmap[ystart][xstart] := 0; end;
       end;

    drawMap;
  end;
end;

//corrects a selection rectangle to encompass specials
procedure TMDIChild.fixsel();
var
  t, x, y, xb, yb : integer;
  changed : boolean;
begin
  Mainform.selRotCCW.Enabled := false;
  Mainform.selRotCW.Enabled := false;


  //makes Xs, Ys the topleft and Xe Ye the bottom right
  if Xe < Xs then
  begin
    t := Xe;
    Xe := Xs;
    Xs := t;
  end;
  if Ye < Ys then
  begin
    t := Ye;
    Ye := Ys;
    Ys := t;
  end;

  //makes sure any specials are fully encompased in the selection

  changed := true;
  while changed do
  begin
    changed := false;
    xb := Xs - 5;
    yb := Ys - 5;
    if xb < 0 then xb := 0;
    if yb < 0 then yb := 0;

    //test for all 3
    for y := yb to ys do
      for x := xb to xe do
      begin
        if fullMap[y][x] = 219 then
        begin
          if y < Ys then begin Ys := y; changed := true; end;
          if x < Xs then begin Xs := x; changed := true; end;
        end
        else if fullMap[y][x] = 220 then
        begin
        if (((ys - y) <= 4) and ((xs - x) <= 4)) then
        begin
          if y < Ys then begin Ys := y; changed := true; end;
          if x < Xs then begin Xs := x; changed := true;
          end;
        end;
        end
        else if fullMap[y][x] = 217 then
        begin
        if (((ys - y) <= 1) and ((xs - x) <= 1)) then
        begin
          if y < Ys then begin Ys := y; changed := true; end;
          if x < Xs then begin Xs := x; changed := true; end;
        end;
        end;

      end;
    for y := yb to ye do
      for x := xb to xs do
      begin
        if fullMap[y][x] = 219 then
        begin
          if y < Ys then begin Ys := y; changed := true; end;
          if x < Xs then begin Xs := x; changed := true; end;
        end
        else if fullMap[y][x] = 220 then
        begin
        if (((ys - y) <= 4) and ((xs - x) <= 4)) then
        begin
          if y < Ys then begin Ys := y; changed := true; end;
          if x < Xs then begin Xs := x; changed := true;
          end;
        end;
        end
        else if fullMap[y][x] = 217 then
        begin
        if (((ys - y) <= 1) and ((xs - x) <= 1)) then
        begin
          if y < Ys then begin Ys := y; changed := true; end;
          if x < Xs then begin Xs := x; changed := true; end;
        end;
        end;

      end;

    for y := Ys to Ye do
      for x := Xs to Xe do
          case fullmap[y][x] of
            217 : begin
              if (y+1) > ye then begin ye := y+1; changed := true; end;
              if (x+1) > xe then begin xe := x+1; changed := true; end;
            end;
            219 : begin
              if (y+5) > ye then begin ye := y+5; changed := true; end;
              if (x+5) > xe then begin xe := x+5; changed := true; end;
            end;
            220 : begin
              if (y+4) > ye then begin ye := y+4; changed := true; end;
              if (x+4) > xe then begin xe := x+4; changed := true; end;
            end;
          end;
  end;

  if (xe-xs) = (ye-ys) then
  begin
    Mainform.selRotCCW.Enabled := true;
    Mainform.selRotCW.Enabled := true;
  end;

end;

//replace left with right tiles
procedure TMDIChild.RepLRClick(Sender: TObject);
var
  x, y : integer;
begin

  newUndo();
  if not selDraw then
  begin
  for y := 0 to MAP_Y - 1 do
    for x := 0 to MAP_X - 1 do
      if fullmap[y][x] = curTileL then
      begin
        addUndo(x,y,curTileL);
        fullMap[y][x] := curTileR;
      end;
  end else
  //only in the selection
  begin
  for y := Ys to Ye do
    for x := Xs to Xe do
      if fullmap[y][x] = curTileL then
      begin
        addUndo(x,y,curTileL);
        fullMap[y][x] := curTileR;
      end;

  end;
end;

//replace right with left tiles
procedure TMDIChild.RepRLClick(Sender: TObject);
var
  x, y : integer;
begin

  newUndo();

  if not selDraw then
  begin
  for y := 0 to MAP_Y - 1 do
    for x := 0 to MAP_X - 1 do
      if fullmap[y][x] = curTileR then
      begin
        addUndo(x,y,curTileR);
        fullMap[y][x] := curTileL;
      end;
  end else
  //only in the selection
  begin
  for y := Ys to Ye do
    for x := Xs to Xe do
      if fullmap[y][x] = curTileR then
      begin
        addUndo(x,y,curTileR);
        fullMap[y][x] := curTileL;
      end;

  end;
end;

//obvious
procedure TMDIChild.EraseMapClick(Sender: TObject);
var
  x, y : integer;
begin

  newUndo();
  for y := 0 to MAP_Y - 1 do
    for x := 0 to MAP_X - 1 do
    begin
      addUndo(x,y,fullMap[y][x]);
      fullMap[y][x] := 0;
    end;
end;

//catches key presses
procedure TMDIChild.FormKeyPress(Sender: TObject; var Key: Char);
begin

//movement for the special tile window
if TileWindow.steveh = true then
begin
  if Key = 'f' then
    if CurtileL <= 4 then
      CurTileL := 190 - (4 - curTileL)
    else
      CurTileL := CurTileL - 4;
  if Key = 'v' then
    if CurTileL >= 186 then
      CurTileL := 4 - (190-curTileL)
    else
      CurTileL := CurTileL + 4;
  if Key = 'c' then
    CurTileL := CurTileL - 1;
  if Key = 'b' then
    CurTileL := CurTileL + 1;
  //right
  if Key = 'F' then
    if CurTileR <= 4 then
      CurTileR := 190 - (4 - CurTileR)
    else
      CurTileR := CurTileR - 4;
  if Key = 'V' then
    if CurTileR >= 186 then
      CurTileR := 4 - (190-CurTileR)
    else
      CurTileR := CurTileR + 4;
  if Key = 'C' then
    CurTileR := CurTileR - 1;
  if Key = 'B' then
    CurTileR := CurTileR + 1;
end
else   //movement for normal tile window
begin
  //left
  if Key = 'f' then
    if CurtileL <= 19 then
      CurTileL := 190 - (19 - curTileL)
    else
      CurTileL := CurTileL - 19;
  if Key = 'v' then
    if CurTileL >= 171 then
      CurTileL := 19 - (190-curTileL)
    else
      CurTileL := CurTileL + 19;
  if Key = 'c' then
    CurTileL := CurTileL - 1;
  if Key = 'b' then
    CurTileL := CurTileL + 1;
  //right
  if Key = 'F' then
    if CurTileR <= 19 then
      CurTileR := 190 - (19 - CurTileR)
    else
      CurTileR := CurTileR - 19;
  if Key = 'V' then
    if CurTileR >= 171 then
      CurTileR := 19 - (190-CurTileR)
    else
      CurTileR := CurTileR + 19;
  if Key = 'C' then
    CurTileR := CurTileR - 1;
  if Key = 'B' then
    CurTileR := CurTileR + 1;
end;
//wrapping code
  if CurTileL = 0 then curTileL := 190;
  if CurTileL = 191 then curTileL := 1;
  if CurTileR = 0 then CurTileR := 190;
  if CurTileR = 191 then CurTileR := 1;

  tileWindow.Invalidate;
end;


//circle and ellipse drawing this gets complex and i don't feel like commenting it
procedure TMDIChild.DrawCircle(const add : boolean);
var
  x, y, p, px, py, Rx, Ry, Rx2, Ry2, twoRx2, twoRy2, cenX, cenY : integer;
begin
  Rx := (abs(Xs - Xe) shr 1);
  Ry := (abs(Ys - Ye) shr 1);
  Rx2 := Rx*Rx;
  Ry2 := Ry*Ry;
  twoRx2 := Rx2 shl 1;
  twoRy2 := Ry2 shl 1;
  x := 0;
  y := Ry;
  px := 0;
  py := twoRx2 * y;

  if Xs < Xe then cenX := Xs + Rx
  else cenX := Xs - Rx;

  if Ys < Ye then cenY := Ys + Ry
  else cenY := Ys - Ry;

  circlePoints(cenX, cenY, x, y, add);
  p := (Ry2 - (Rx2 * Ry) + (Rx2 shr 4));

  while px < py do
  begin
    inc(x);
    inc(px,twoRy2);
    if p < 0 then
      inc(p,(Ry2 + px))
    else begin
      dec(y);
      dec(py,twoRx2);
      inc(p,(ry2+px-py));
    end;
    circlePoints(cenX, cenY, x, y, add);
  end;

  p := round( Ry2 * (x+0.5)*(x+0.5) + Rx2*(y-1)*(y-1) -Rx2*Ry2);
  while y > 0 do
  begin
    dec(y);
    dec(py,twoRx2);
    if p > 0 then
      inc(p,(Rx2-py))
    else begin
      inc(x);
      inc(px,twoRy2);
      inc(p, (Rx2-py+px));
    end;
    circlePoints(cenX, cenY, x, y, add);
  end;
end;


//more with circles and ellipses not gonna comment it all
procedure TMDIChild.circlePoints(const cx, cy, x, y : integer; const add : boolean);
begin

  if zoom > 3 then
  begin
    if usedTile <> 0 then
    begin
      mapDraw.Surface.Canvas.CopyRect(bounds((cx+x-topLeftX) shl zoom, (cy+y-topLeftY) shl zoom, 1 shl zoom, 1 shl zoom), tile.Canvas,tileRect(usedTile));
      mapDraw.Surface.Canvas.CopyRect(bounds((cx-x-topLeftX) shl zoom, (cy+y-topLeftY) shl zoom, 1 shl zoom, 1 shl zoom), tile.Canvas,tileRect(usedTile));
      mapDraw.Surface.Canvas.CopyRect(bounds((cx+x-topLeftX) shl zoom, (cy-y-topLeftY) shl zoom, 1 shl zoom, 1 shl zoom), tile.Canvas,tileRect(usedTile));
      mapDraw.Surface.Canvas.CopyRect(bounds((cx-x-topLeftX) shl zoom, (cy-y-topLeftY) shl zoom, 1 shl zoom, 1 shl zoom), tile.Canvas,tileRect(usedTile));
    end else begin
      mapDraw.Surface.Canvas.Brush.Color := clBlack;
      mapDraw.Surface.Canvas.FillRect(bounds((cx+x-topLeftX) shl zoom, (cy+y-topLeftY) shl zoom, 1 shl zoom, 1 shl zoom));
      mapDraw.Surface.Canvas.FillRect(bounds((cx-x-topLeftX) shl zoom, (cy+y-topLeftY) shl zoom, 1 shl zoom, 1 shl zoom));
      mapDraw.Surface.Canvas.FillRect(bounds((cx+x-topLeftX) shl zoom, (cy-y-topLeftY) shl zoom, 1 shl zoom, 1 shl zoom));
      mapDraw.Surface.Canvas.FillRect(bounds((cx-x-topLeftX) shl zoom, (cy-y-topLeftY) shl zoom, 1 shl zoom, 1 shl zoom));
    end;
  end else begin
    case usedTile of
      0: mapDraw.Surface.Canvas.Brush.Color := clBlack;
      1..161: mapDraw.Surface.Canvas.Brush.Color := clSilver;
      162..169: mapDraw.Surface.Canvas.Brush.Color := clBlue;
      170: mapDraw.Surface.Canvas.Brush.Color := clYellow;
      171: mapDraw.Surface.Canvas.Brush.Color := clLime;
      172: mapDraw.Surface.Canvas.Brush.Color := clMaroon;
      173..190: mapDraw.Surface.Canvas.Brush.Color := clAqua;
    end;
    mapDraw.Surface.Canvas.FillRect(bounds((cx+x-topLeftX) shl zoom, (cy+y-topLeftY) shl zoom, 1 shl zoom, 1 shl zoom));
    mapDraw.Surface.Canvas.FillRect(bounds((cx-x-topLeftX) shl zoom, (cy+y-topLeftY) shl zoom, 1 shl zoom, 1 shl zoom));
    mapDraw.Surface.Canvas.FillRect(bounds((cx+x-topLeftX) shl zoom, (cy-y-topLeftY) shl zoom, 1 shl zoom, 1 shl zoom));
    mapDraw.Surface.Canvas.FillRect(bounds((cx-x-topLeftX) shl zoom, (cy-y-topLeftY) shl zoom, 1 shl zoom, 1 shl zoom));
  end;
  if add then
  begin
    placeTile(cx+x, cy+y, usedTile);
    placeTile(cx+x, cy-y, usedTile);
    placeTile(cx-x, cy+y, usedTile);
    placeTile(cx-x, cy-y, usedTile);
  end;
end;


{
filling an area paint bucket style
basic concept takes the starting point tests right if matching then it will move there and start testing again.
tests go around right down left up
if all tests fail then the tile is changed and it returns to the previous one.
returns true when it reaches an end point
}
function TMDIChild.fillArea(const x, y : smallint; const old : byte) : boolean;
begin

//if trying to fill on the clicked tile just exit
if usedTile = old then
  begin result := true; exit; end;
//if you run off the map
if ((x < 0) or (x > 1023)) then
  begin result := true; exit; end;
if ((y < 0) or (y > 1023)) then
  begin result := true; exit; end;

placeTile(x,y,usedTile);

fillRight(x+1,y,old);
fillDown(x,y+1,old);
fillLeft(x-1,y,old);
fillUp(x,y-1,old);

result := true;
end;

procedure TMDIChild.fillRight(const x, y : smallint; const old : byte);
begin
  if x > 1023 then exit;
  if fullmap[y][x] = old then
  begin
    placeTile(x,y,usedTile);
    fillRight(x+1,y,old);
    fillDown(x,y+1,old);
    fillUp(x,y-1,old);
  end;
end;

procedure TMDIChild.fillDown(const x, y : smallint; const old : byte);
begin
  if y > 1023 then exit;
  if fullmap[y][x] = old then
  begin
    placeTile(x,y,usedTile);
    fillRight(x+1,y,old);
    fillDown(x,y+1,old);
    fillLeft(x-1,y,old);
  end;
end;

procedure TMDIChild.fillLeft(const x, y : smallint; const old : byte);
begin
  if x < 0 then exit;
  if fullmap[y][x] = old then
  begin
    placeTile(x,y,usedTile);
    fillDown(x,y+1,old);
    fillLeft(x-1,y,old);
    fillUp(x,y-1,old);
  end;
end;

procedure TMDIChild.fillUp(const x, y : smallint; const old : byte);
begin
  if y < 0 then exit;
  if fullmap[y][x] = old then
  begin
    placeTile(x,y,usedTile);
    fillRight(x+1,y,old);
    fillLeft(x-1,y,old);
    fillUp(x,y-1,old);
  end;
end;

procedure TMDIChild.actEditClearExecute(Sender: TObject);
var
  x, y : integer;
begin

  newUndo();
  if selDraw then
  begin
    for y := Ys to Ye do
      for x := Xs to Xe do
        placeTile(x,y, 0);

  drawMap;
  end;

end;

procedure TMDIChild.actEditUndoExecute(Sender: TObject);
var
  a : integer;
begin
  if length(undo[0].x) = 0 then exit;
  newRedo();
  for a := 0 to high(undo[0].x) do
  begin
    addRedo(undo[0].x[a],undo[0].y[a],fullMap[undo[0].y[a]][undo[0].x[a]]);
    fullMap[undo[0].y[a]][undo[0].x[a]] := undo[0].tile[a];
  end;

  for a := 0 to high(undo)-1 do
   undo[a] := undo[a+1];

  setLength(undo[high(undo)].x,0);
  setLength(undo[high(undo)].y,0);
  setLength(undo[high(undo)].tile,0);

  drawMap;
end;

procedure TMDIChild.addUndo(const xPos,yPos : integer; const tileb : byte);
begin
  setLength(undo[0].x,length(undo[0].x)+1);
  setLength(undo[0].y,length(undo[0].y)+1);
  setLength(undo[0].tile,length(undo[0].tile)+1);
  undo[0].x[high(undo[0].x)] := xPos;
  undo[0].y[high(undo[0].y)] := yPos;
  undo[0].tile[high(undo[0].tile)] := tileb;
end;

procedure TMDIChild.newUndo();
var
  a : integer;
begin
  a := length(undo)-1;
  while a > 0 do
  begin
    undo[a] := undo[a-1];
    dec(a);
  end;

  setLength(undo[0].x,0);
  setLength(undo[0].y,0);
  setLength(undo[0].tile,0);
end;

procedure TMDIChild.newRedo();
var
  a : integer;
begin
  a := length(redo)-1;
  while a > 0 do
  begin
    redo[a] := redo[a-1];
    dec(a);
  end;

  setLength(redo[0].x,0);
  setLength(redo[0].y,0);
  setLength(redo[0].tile,0);
end;

procedure TMDIChild.addRedo(const xPos,yPos : integer; const tileb : byte);
begin
  setLength(redo[0].x,length(redo[0].x)+1);
  setLength(redo[0].y,length(redo[0].y)+1);
  setLength(redo[0].tile,length(redo[0].tile)+1);
  redo[0].x[high(redo[0].x)] := xPos;
  redo[0].y[high(redo[0].y)] := yPos;
  redo[0].tile[high(redo[0].tile)] := tileb;
end;

procedure TMDIChild.actEditRedoExecute(Sender: TObject);
var
  a : integer;
begin
  if Length(redo[0].x) = 0 then exit;
  newUndo();
  for a := 0 to high(redo[0].x) do
  begin
    addUndo(redo[0].x[a],redo[0].y[a],fullMap[redo[0].y[a]][redo[0].x[a]]);
    fullMap[redo[0].y[a]][redo[0].x[a]] := redo[0].tile[a];
  end;

  for a := 0 to high(redo)-1 do
   redo[a] := redo[a+1];

  setLength(redo[high(redo)].x,0);
  setLength(redo[high(redo)].y,0);
  setLength(redo[high(redo)].tile,0);

  drawMap;
end;

procedure TMDIChild.WMDROPFILES (var Msg: TMessage);
var
 i, j,
 amount,
 size: integer;
 Filename: PChar;
 S: string;
 begin
 inherited;
 Filename := '';
 Amount := DragQueryFile(Msg.WParam, $FFFFFFFF, Filename, 255);
 for i := 0 to (Amount - 1) do
 begin
  size := DragQueryFile(Msg.WParam, i , nil, 0) + 1;
  Filename:= StrAlloc(size);
  DragQueryFile(Msg.WParam,i , Filename, size);
  s := Extractfileext(Filename);
 if AnsiLowerCase(s) = '.lvl' then
 begin
   if MainForm.MDIChildCount > 0 then
    for j := 0 to MainForm.MDIChildCount-1 do
    begin
      if ((MainForm.MDIChildren[j] as TMDIChild).getFileName() = FileName) then
      begin
        MainForm.MDIChildren[j].BringToFront;
        exit;
      end;
    end;

    MainForm.CreateMDIChild(FileName);
    //MainForm.MRU.AddHistoryFile(FileName); //removed until reinstalled

 end
 else if AnsiLowerCase(s) <> '.bmp' then
  exit
 else
 begin
   loadTiles(filename);
 end;
  StrDispose(Filename);
 end;
 DragFinish(Msg.WParam);
end;

procedure TMDIChild.actEditCutExecute(Sender: TObject);
var
  x, y : word;
begin
  setLength(MainForm.clip,Ye - Ys + 1, Xe - Xs + 1);

  newundo();
  for y := 0 to (Ye-Ys) do
    for x := 0 to (Xe-Xs) do
    begin
      MainForm.clip[y][x] := fullMap[Ys + y][Xs + x];
      addUndo(Xs + x, Ys+y,fullMap[Ys + y][Xs + x]);
      fullMap[Ys + y][Xs + x] := 0;
    end;
  drawMap;
end;

//floats the selection ie removes it from the map
procedure TMDIChild.floatsel(Shift: TShiftState);
var
  x, y: word;
begin
  setLength(float,Ye - Ys + 1, Xe - Xs + 1);
  fX := Xs;
  fY := Ys;
  for y := 0 to (ye-ys) do
    for x := 0 to (xe-xs) do
    begin
      float[y][x] := fullmap[ys + y][xs+x];
      if not (ssCtrl in Shift) then  //if ctrl is held it floats a copy
        fullmap[ys + y][xs+x] := 0;
    end;

end;

//defloats a selection, puts it back down on the map.
procedure TMDIChild.defloatsel(Shift: TShiftState);
var
  x, y : word;
begin

  setLength(float,0,0);
  hasfloat := false;
end;

procedure TMDIChild.setStatus();
var
 status : String;
begin
  status := inttostr(TopLeftX) + ',' + inttostr(TopLeftY) + ' - ';
  status := status + inttostr(TopLeftX+dispX-1) + ',' + inttostr(TopLeftY+dispY-1) + ' :: ';
  MainForm.StatusBar.Panels.Items[2].Text := status;
end;

procedure TMDIChild.rotateCCW();
var
  rot : array of array of byte;
  x,y : word;
begin
  if selDraw then
  begin
    setLength(rot, (Ye - Ys + 1), (Xe - Xs + 1));

    for y := 0 to ye-ys do
      for x := 0 to xe-xs do
        case fullmap[ys+x][xe-y] of
          1..190,216,218: rot[y][x] := fullmap[ys+x][xe-y];
          217: rot[y-1][x] := fullmap[ys+x][xe-y];
          219: rot[y-5][x] := fullmap[ys+x][xe-y];
          220: rot[y-4][x] := fullmap[ys+x][xe-y];
        end;
    for y := 0 to ye-ys do
      for x := 0 to xe-xs do
        fullmap[ys+y][xs+x] := rot[y][x];

    drawMap;
  end;
end;

procedure TMDIChild.rotateCW();
var
  rot : array of array of byte;
  x,y : word;
begin
  if selDraw then
  begin
    setLength(rot, (Ye - Ys + 1), (Xe - Xs + 1));

    for y := 0 to ye-ys do
      for x := 0 to xe-xs do
        case fullmap[ye-x][xs+y] of
          1..190,216,218: rot[y][x] := fullmap[ye-x][xs+y];
          217: rot[y][x-1] := fullmap[ye-x][xs+y];
          219: rot[y][x-5] := fullmap[ye-x][xs+y];
          220: rot[y][x-4] := fullmap[ye-x][xs+y];
        end;
    for y := 0 to ye-ys do
      for x := 0 to xe-xs do
        fullmap[ys+y][xs+x] := rot[y][x];

    drawMap;
  end;
end;


end.
