unit RADAR;

interface

uses
  Windows, Messages, Variants, Classes, Graphics, Controls, DxClass,
  StdCtrls, DXDraws, ExtCtrls, ComCtrls, CHILDWIN;

type
  TRadarBox = class(TDxForm)
    RadarDisp: TDXDraw;
    Hori: TScrollBar;
    Vert: TScrollBar;
    pnlSpace: TPanel;
    pnlBottom: TPanel;
    procedure FormResize(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure RadarDispMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure HoriChange(Sender: TObject);
    procedure VertChange(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    { Private declarations }
    radX : word;
    radY : word;
    showX : word;
    showY : word;
    difX : byte;
    difY : byte;
    child : TMDIChild;
  public
    { Public declarations }
    procedure setX(newx : word);
    procedure setY(newy : word);
    procedure setChild(c : TMDIChild);
    procedure DrawRadar();
  end;

var
  RadarBox: TRadarBox;

implementation

uses MAIN;

{$R *.dfm}

//should be given as center of screen
procedure TRadarBox.SetX(newx : word);
begin
  if newx < (showX shr 1) then
    radX := 0
  else if newx > (1023 - (showX shr 1)) then
    radX := 1023 - showX
  else
    radX := newx - (showX shr 1);

  Hori.Position := RadX;
end;

procedure TRadarBox.SetY(newy : word);
begin
  if newy < (showY shr 1) then
    radY := 0
  else if newy > (1023 - (showY shr 1)) then
    radY := 1023 - showY
  else
    radY := newy-(showY shr 1);

  Vert.Position := RadY;
end;

procedure TRadarBox.setChild(c : TMDIChild);
begin
  child := c;
end;

procedure TRadarBox.FormResize(Sender: TObject);
begin
  if RadarDisp.Height > 1022 then begin
    vert.Max := 1023;
    showY := 1023;
  end
  else begin
    vert.Max := 1023 - RadarDisp.Height;
    showY := RadarDisp.Height;
  end;

  if RadarDisp.Width > 1022 then begin
    hori.Max := 1023;
    showX := 1023;
  end
  else begin
    hori.Max := 1023 - RadarDisp.Width;
    showX := RadarDisp.Width;
  end;

  vert.LargeChange := showY-1;
  hori.LargeChange := showX-1;
  vert.Position := radY;
  hori.Position := radX;

  DrawRadar();
end;

procedure TRadarBox.DrawRadar();
var
  curX, curY, sX, sY : integer;
  curIndex : byte;
begin
  if not radarDisp.CanDraw then exit;
  if child = nil then exit;

  radarDisp.Surface.Fill(0);


  //makes sure the red screen box is drawn as often as possible
  if radX > Child.dispX then sX := -child.dispX else sX := 0;
  if radY > Child.dispY then sY := -child.dispY else sY := 0;

  with RadarDisp.Surface.Canvas do
  begin

    for curY := sY to showY do
      for curX := sX to showX  do
      begin

        
        curIndex := Child.fullMap[curY+radY][curX+radX];
        moveTo(curX, curY);
        case curIndex of
          //0: begin pen.Color := clred; lineTo(curX+1, curY+1); end;
          1..161: Pixels[curX, curY] :=  clSilver;
          162..169: Pixels[curX, curY] :=  clBlue;
          170: Pixels[curX, curY] := clYellow;
          171: Pixels[curX, curY] :=  clLime;
          172: Pixels[curX, curY] := clMaroon;
          173..190: Pixels[curX, curY] :=  clAqua;
          216, 218: Pixels[curX, curY] := clGray;
          217:
            begin
              Brush.Color := clGray;
              fillRect(bounds(curX, curY, 2, 2));
            end;
          219:
            begin
              Brush.Color := clRed;
              fillRect(Bounds(curX, curY, 6, 6));
            end;
          220:
            begin
              Brush.Color := clFuchsia;
              fillRect(Bounds(curX, curY, 5, 5));
            end;
        end;

        //draws the red bounding box in the radar that represents the screen
        {if ((curY + radY) = Child.topLeftY) then
          if ((curX + radX) = Child.topLeftX) then
            begin
              Brush.Color := clRed;
              frameRect(bounds(curX, curY, Child.dispX, Child.dispY));
            end;    }
      end;

    Brush.Color := clRed;
    frameRect(bounds(child.topleftX - radX, child.topleftY - radY,Child.dispX, Child.dispY));

    Release;
  end;
  radarDisp.Flip;
end;

procedure TRadarBox.FormCreate(Sender: TObject);
begin
  radX := 0;
  radY := 0;

  difX := Width - ClientWidth;
  difY := Height - ClientHeight;
  Width := 200+difX;
  Height := 200+difY;
  Left := (MainForm.Left + MainForm.Width) - Width;
  Top := (MainForm.Top + MainForm.Height) - Height;
  showX := RadarDisp.Width;
  showY := RadarDisp.Height;
end;

procedure TRadarBox.radarDispMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  xLoc, yLoc : word;
begin
  if Button = mbLeft then
  begin
    xLoc := X+radX;
    yLoc := Y+radY;
    if xLoc < (Child.dispX shr 1) then
      Child.topLeftX := 0
    else if xLoc > (1023 - Child.dispX) then
      Child.topLeftX := 1023 - Child.dispX
    else
      Child.topLeftX := xLoc - (Child.dispX shr 1);

    if yLoc < (Child.dispY shr 1) then
      Child.topLeftY := 0
    else if yLoc > (1023 - Child.dispY) then
      Child.topLeftY := 1023 - Child.dispY
    else
      Child.topLeftY := yLoc - (Child.dispY shr 1);

  //makes the radar recenter..to change this comment this out or put in a check
  setX(xLoc);
  setY(yLoc);

  vert.Position := radY;
  hori.Position := radX;

  Child.mapVertScroll.Position := Child.topLeftY;
  Child.mapHoriScroll.Position := Child.topLeftX;


  drawRadar();
  Child.drawMap();
  end;
end;

procedure TRadarBox.HoriChange(Sender: TObject);
begin
  if showX <> 1023 then
  begin
    radX := hori.Position;
    DrawRadar;
  end;
end;

procedure TRadarBox.VertChange(Sender: TObject);
begin
  if showY <> 1023 then
  begin
    radY := vert.Position;
    DrawRadar;
  end;
end;

procedure TRadarBox.FormHide(Sender: TObject);
begin
  Mainform.ViewRadar1.Checked := false;
end;

procedure TRadarBox.FormShow(Sender: TObject);
begin
  MainForm.ViewRadar1.Checked := true;
end;

end.
