unit TILE;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, DXDraws, CHILDWIN, StdCtrls;

const
  TILE_SIZE = 16;
  TILE_SHIFT = 4;
  TILE_COUNT = 190;
  TILE_COLUMNS = 19;
  TILE_ROWS = 10;
  TILE_WIDTH = 304;
  TILE_HEIGHT = 160;

type
  TTileWindow = class(TForm)
    scroll: TScrollBar;
    procedure FormCreate(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormPaint(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure scrollChange(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    steveh : boolean;
    procedure toggleSteveh(b : boolean);
  end;

var
  TileWindow: TTileWindow;

implementation

uses MAIN, TILESELECTED;

{$R *.dfm}

function tileRect(index: word) : TRect;
begin
//faster method?
  case index of
    1..19: result:=Bounds((index-1) shl 4, 0, TILE_SIZE,TILE_SIZE);
    20..38: result:=Bounds(((index-1)-19) shl 4, 16, TILE_SIZE,TILE_SIZE);
    39..57: result:=Bounds(((index-1)-38) shl 4, 32, TILE_SIZE,TILE_SIZE);
    58..76: result:=Bounds(((index-1)-57) shl 4, 48, TILE_SIZE,TILE_SIZE);
    77..95: result:=Bounds(((index-1)-76) shl 4, 64, TILE_SIZE,TILE_SIZE);
    96..114: result:=Bounds(((index-1)-95) shl 4, 80, TILE_SIZE,TILE_SIZE);
    115..133: result:=Bounds(((index-1)-114) shl 4, 96, TILE_SIZE,TILE_SIZE);
    134..152: result:=Bounds(((index-1)-133) shl 4, 112, TILE_SIZE,TILE_SIZE);
    153..171: result:=Bounds(((index-1)-152) shl 4, 128, TILE_SIZE,TILE_SIZE);
    172..190: result:=Bounds(((index-1)-171) shl 4, 144, TILE_SIZE,TILE_SIZE);
  end;
end;

procedure TTileWindow.toggleSteveh(b : boolean);
var
  heightdif : word;
begin
  if b then //turn steveh mode on
  begin
    BorderStyle := bsSizeToolWin;
    clientHeight := 160;
    scroll.Visible := true;
    heightdif := Height - ClientHeight;
    ClientWidth := scroll.Width + 65;
    Constraints.MinWidth := Width;
    Constraints.MaxWidth := Width;
    Constraints.MinHeight := Height;
    Constraints.MaxHeight := 768 + Heightdif;
    scroll.Position := 0;
    steveh := true;
  end
  else
  begin
    BorderStyle := bsToolWindow;
    Constraints.MinHeight := 0;
    Constraints.MinWidth := 0;
    Constraints.MaxHeight := 0;
    Constraints.MaxWidth := 0;
    clientWidth := 304;
    clientHeight := 160;
    scroll.Visible := false;
    steveh := false;
  end;

end;

procedure TTileWindow.FormCreate(Sender: TObject);
begin

  top := (MainForm.Top + MainForm.Height) - Height;
  Left := MainForm.Left;
  clientWidth := 304;
  clientHeight := 160;
  steveh := false;
  toggleSteveh(steveh);
end;


procedure TTileWindow.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  tempIndex: word;
begin
  //figuring out which tile was actually clicked from the mouse x and y values
  if steveh then
  begin
    tempIndex := ((scroll.Position shl 2)) + (Y shr TILE_SHIFT) shl 2;
    inc(tempIndex, (X shr TILE_SHIFT));
  end
  else
  begin
    tempIndex := TILE_COLUMNS * (Y shr TILE_SHIFT);
    inc(tempIndex,(X shr TILE_SHIFT));
  end;
  //which button was pressed
  case Button of
    mbLeft: if (Mainform.MDIChildCount > 0) then
              (MainForm.ActiveMDIChild as TMDIChild).setLeft(tempIndex+1);
    mbRight: if (Mainform.MDIChildCount > 0) then
              (MainForm.ActiveMDIChild as TMDIChild).setRight(tempIndex+1);
  end;
  Invalidate;
  TileSel.Invalidate;
  if Mainform.Tool in [1..8] then
  begin
    MainForm.Tool := 0;
    if Mainform.MDIChildCount > 0  then
    for tempIndex := 0 to MainForm.MDIChildCount - 1 do
      (MainForm.MDIChildren[tempindex] as TMDIChild).mapDraw.Cursor := 16;
    mainForm.SetPen.Checked := true;
  end;

end;

procedure TTileWindow.FormPaint(Sender: TObject);
var
  L,R : byte;
  a, b, tile : byte;
begin
    if (Mainform.MDIChildCount < 1) then exit;
  L := (MainForm.ActiveMDIChild as TMDIChild).getLeft();
  R := (MainForm.ActiveMDIChild as TMDIChild).getRight();

  with Canvas do
  begin
  if Steveh then
    begin
      tile := (scroll.Position shl 2) + 1;
      for b := 0 to (ClientHeight shr TILE_SHIFT) do
        for a := 0 to 3 do
          begin
            if tile <= 190 then
            CopyRect(bounds(a shl TILE_SHIFT, b shl TILE_SHIFT, TILE_SIZE, TILE_SIZE),
                     (MainForm.ActiveMDIChild as TMDIChild).tile.Canvas,
                     tileRect(tile));

            if ((L = R) and (L = tile)) then
            begin
              Brush.Color := clWhite;
              FrameRect(bounds(a shl TILE_SHIFT, b shl TILE_SHIFT, TILE_SIZE, TILE_SIZE));
            end
            else if tile = L then
            begin
              Brush.Color := clYellow;
              FrameRect(bounds(a shl TILE_SHIFT, b shl TILE_SHIFT, TILE_SIZE, TILE_SIZE));
            end
            else if Tile = R then
            begin
              Brush.Color := clRed;
              FrameRect(bounds(a shl TILE_SHIFT, b shl TILE_SHIFT, TILE_SIZE, TILE_SIZE));
            end;

            inc(tile);
          end;

    end
    else
    begin
      Draw(0,0,(MainForm.ActiveMDIChild as TMDIChild).tile); //copy the tiles over from the tile TBitmap

      //drawing the red and yellow or white box around the tiles
      if L <> R then
      begin
          Brush.Color := clYellow;
          FrameRect(tileRect(L));
          Brush.Color := clRed;
          FrameRect(tileRect(R));
      end
      else
      begin
        Brush.Color := clWhite;
        FrameRect(tileRect(L));
      end;
    end;
  end; //end with
end;

procedure TTileWindow.FormShow(Sender: TObject);
begin
  MainForm.ViewTileWindow.Checked := true;
end;

procedure TTileWindow.FormHide(Sender: TObject);
begin
  MainForm.ViewTileWindow.Checked := false;
end;

procedure TTileWindow.FormResize(Sender: TObject);
begin
  clientHeight := (clientHeight shr TILE_SHIFT) shl TILE_SHIFT;
  if (clientHeight shr TILE_SHIFT) >= 48 then
  begin
    scroll.Max := 0;
    scroll.position := 0;
  end
  else
  begin
    scroll.Max := (48 - (clientHeight shr TILE_SHIFT));
    if scroll.Position > scroll.Max then
      scroll.Position := scroll.Max;
  end;

end;

procedure TTileWindow.scrollChange(Sender: TObject);
begin
mainform.StatusBar.SimpleText := IntToStr(scroll.position);
  invalidate;
end;

end.
